#!/usr/bin/python3

'''
Copyright (c) 2021 Eric D. Cohen

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

import sys
import json
import random
import string
import argparse
from pathlib import Path
from faker import Faker

class DictGen:
    '''
    Random dictionary generator
    '''
    def __init__(self, seed=7):
        self.fake = Faker()
        random.seed(seed)
        Faker.seed(seed)

    @staticmethod
    def _gen_rand_str(maxlen):
        chars = string.ascii_letters + string.digits + '-_ '
        return lambda: ''.join(random.choices(chars, k=random.randint(3, maxlen)))

    def gen_dict(self, breadth_rng, depth_rng, list_dist=(1, 1), leaf_fcn=None,
                 inner_fcn=None, curr_depth=0, gen_out=None):
        '''
        Generic random nested dictionary generator. Includes nested lists.
        '''
        if gen_out is None:
            gen_out = {}
        if not leaf_fcn:
            leaf_fcn = self._gen_rand_str(24)
        if not inner_fcn:
            inner_fcn = self._gen_rand_str(8)

        for _ in range(random.randint(*breadth_rng)):
            if curr_depth < random.randint(max(curr_depth, depth_rng[0]), depth_rng[1]):
                new_val = {} if random.choices((True, False), list_dist)[0] else []
                if isinstance(gen_out, dict):
                    k = inner_fcn()
                    gen_out[k] = new_val
                    self.gen_dict(breadth_rng, depth_rng, list_dist, leaf_fcn, inner_fcn,
                                  curr_depth + 1, gen_out[k])
                else:
                    assert isinstance(gen_out, list)
                    gen_out.append(new_val)
                    self.gen_dict(breadth_rng, depth_rng, list_dist, leaf_fcn, inner_fcn,
                                  curr_depth + 1, gen_out[-1])
            else:
                if isinstance(gen_out, dict):
                    gen_out[inner_fcn()] = leaf_fcn()
                else:
                    assert isinstance(gen_out, list)
                    gen_out.append(leaf_fcn())

        return gen_out

    def gen_fake_dict(self, breadth_rng=(2, 3), depth_rng=(2, 4), list_dist=(1, 1)):
        '''
        Generate random dictionary using various faker methods
        '''
        ngen = lambda: None
        fgen = lambda: random.gauss(0, 2**20)
        igen = lambda: round(random.gauss(0, 2**20))
        leaf_choices = (self.fake.street_address, self.fake.company,
                        random.random, ngen, fgen, igen)
        leaf_fcn = lambda: random.choice(leaf_choices)()

        inner_choices = (self.fake.word,)
        inner_fcn = lambda: random.choice(inner_choices)()

        return self.gen_dict(breadth_rng, depth_rng, list_dist, inner_fcn=inner_fcn,
                             leaf_fcn=leaf_fcn)

def main():
    ''' CLI entry point '''
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument('-a', dest='args', required=False,
                            action='store_true', default=False,
                            help='Include args in output dictionary')
    arg_parser.add_argument('-b', dest='breadth_min', required=False,
                            default=2, type=int,
                            help='Minimum breadth')
    arg_parser.add_argument('-B', dest='breadth_max', required=False,
                            default=3, type=int,
                            help='Maximum breadth')
    arg_parser.add_argument('-c', dest='count', required=False,
                            default=1, type=int,
                            help='Number of JSON files to generate')
    arg_parser.add_argument('-d', dest='depth_min', required=False,
                            default=2, type=int,
                            help='Minimum depth')
    arg_parser.add_argument('-D', dest='depth_max', required=False,
                            default=4, type=int,
                            help='Maximum depth')
    arg_parser.add_argument('-i', dest='indent', required=False,
                            default=2, type=int,
                            help='Output indent level (0 for compact)')
    arg_parser.add_argument('-l', dest='pct_lists', required=False,
                            default=50, type=int,
                            help='Percent nested lists')
    arg_parser.add_argument('-o', dest='outfile', required=False,
                            default=None, type=Path,
                            help='Output file')
    arg_parser.add_argument('-s', dest='seed', required=False,
                            default=7, type=int,
                            help='PRNG seed')
    args = arg_parser.parse_args()

    breadth = (args.breadth_min, args.breadth_max)
    if breadth[0] >= breadth[1]:
        print("-b value must be less than -B")
        sys.exit(1)

    depth = (args.depth_min, args.depth_max)
    if depth[0] >= depth[1]:
        print("-d value must be less than -D")
        sys.exit(1)

    if not 0 <= args.pct_lists < 100:
        print("-l value must be in the range [0, 100)")
        sys.exit(1)
    pct_lists = (100 - args.pct_lists, args.pct_lists)

    if args.outfile:
        outfile = args.outfile
    else:
        outfile = Path('/dev/stdout')

    if args.count == 1:
        outfiles = [outfile]
    else:
        mkfn = lambda f, s: f.parent / Path(f.stem + s).with_suffix(f.suffix)
        outfiles = [mkfn(outfile, '{:03d}'.format(n)) for n in range(args.count)]

    dict_gen = DictGen(args.seed)
    for outfile in outfiles:
        out_dict = dict_gen.gen_fake_dict(breadth, depth, pct_lists)
        if args.args:
            out_dict['__argv__'] = sys.argv
        with open(outfile, 'w') as handle:
            json.dump(out_dict, handle, indent=args.indent if args.indent else None)

if __name__ == "__main__":
    main()



#!/usr/bin/python3

'''
Copyright (c) 2021 Eric D. Cohen

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

import json
import random
import argparse
import collections
from pathlib import Path

# Use sys.stdin/sys.stdout instead...
STDIN = Path('/dev/stdin')
STDOUT = Path('/dev/stdout')

class DictSam:
    '''
    Dictionary split and merge (DICTSAM) main class.
    '''
    ignore_leaves = False
    enforce_unique = False
    def __init__(self, data=None, denormed=False, enforce_unique=False, enforce_serdes=True):
        DictSam.enforce_unique = enforce_unique
        if not isinstance(data, (dict, list, type(None))):
            raise TypeError("Data must contain root of dictionary or list type")
        if denormed:
            self._data = self.normalize(data)
        else:
            self._data = data
        if enforce_serdes:
            # DictSam currently only supports JSON-serializable dictionaries
            # This is a hacky admission control mechanism to guarantee that.
            self._data = json.loads(json.dumps(self._data))

    @classmethod
    def set_ignore_leaves(cls, ignore_leaves):
        ''' Set ignore leaves behavior '''
        DictSam.ignore_leaves = ignore_leaves

    def denormalize(self, data=None):
        '''
        Convert a nested dictionary into a list of lists of paths
        '''
        denormed = []
        def helper(data, curr_path=None):
            if not curr_path:
                curr_path = []
            if isinstance(data, dict):
                for key, value in data.items():
                    helper(value, curr_path + [key])
            elif isinstance(data, list):
                # Convert nested lists to dicts for convenience
                for idx, value in enumerate(data):
                    helper(value, curr_path + [idx])
            else:
                denormed.append(curr_path + [data])
        if data:
            helper(data)
        else:
            helper(self._data)
        return denormed

    @classmethod
    def normalize(cls, data):
        '''
        Convert a list of lists of paths into a nested dictionary
        '''
        nested_dict = lambda: collections.defaultdict(nested_dict)
        cursor = cursor_root = collections.defaultdict(nested_dict)
        for path in data:
            assert len(path) > 1 # Dict consists of pairs
            key = None
            for key in path[:-1]:
                if not isinstance(cursor[key], dict):
                    if cls.enforce_unique:
                        raise RuntimeError('Path "{}" overwrites existing path'
                                           .format('.'.join([str(x) for x in path])))
                    # This is an overwrite of an existing leaf so smash it
                    cursor[key] = collections.defaultdict(nested_dict)
                cursor_prev = cursor
                cursor = cursor[key]
            cursor_prev[key] = path[-1]
            cursor = cursor_root
        return cls._restore_lists(cursor_root)

    def random_dict_pick(self, pct_pick):
        '''
        Returns a DictSam with a random selection of paths.

        pct_pick -- Percentage of paths to randomly select.  Note two calls
        with > 50 must contain some overlapping paths.
        '''

        denorm = self.denormalize()
        num_picks = pct_pick * len(denorm) // 100
        denorm = random.sample(denorm, num_picks)
        return DictSam(denorm, denormed=True)

    def get_data(self):
        ''' Get data dictionary '''
        return self._data

    def items(self):
        ''' Get data dictionary items '''
        return self._data.items()

    @classmethod
    def _restore_lists(cls, pure_dict, mixed_dict=None):
        '''
        Walk a nested dictionary and convert each subdictionary with integer
        keys into nested lists. Note that there is a slight loss of dictionary
        string integer semantics. Also converts defaultdicts back to dicts.

        pure_dict -- Pure nested dictionary (no lists) input
        mixed_dict -- Output dictionary consisting of nested dictionaries and lists
        '''
        if not mixed_dict:
            ints = all([cls._is_int(x) for x in pure_dict.keys()])
            if ints and pure_dict.keys():
                mixed_dict = [None] * (max(pure_dict.keys()) + 1)
            else:
                mixed_dict = {}
        for key, value in pure_dict.items():
            if isinstance(value, dict):
                # Discards list-as-dict semantics
                if all([cls._is_int(x) for x in value.keys()]):
                    # Fill sparse list indices with Nones
                    sorted_by_key_vals = [(k, value.get(k)) for k in range(max(value.keys()) + 1)]
                    mixed_dict[key] = [x[1] for x in sorted_by_key_vals]
                else:
                    mixed_dict[key] = dict(value)
                cls._restore_lists(value, mixed_dict[key])
            else:
                mixed_dict[key] = value
        return mixed_dict

    @staticmethod
    def _is_int(val):
        ''' Utility function to check for an integer string. '''
        try:
            int(val)
            return True
        except ValueError:
            return False

    @classmethod
    def _cmp_list(cls, left, right):
        '''
        Compare two lists element by element
        '''
        end = min(len(left), len(right))
        if DictSam.ignore_leaves:
            end -= 1
        left = left[:end]
        right = right[:end]
        for lelm, relm in zip(left, right):
            if lelm != relm:
                return False
        return True

    @classmethod
    def _cmp_lists(cls, left, right):
        '''
        Check if a list exists among a list of lists

        left - List of interest
        right -- List of lists against which to check
        '''
        assert not isinstance(left[0], list)
        for del_elm in right:
            if cls._cmp_list(left, del_elm):
                return True
        return False

    def _sub_and(self, other, is_or=True):
        data_a_denorm = self.denormalize()
        data_b_denorm = other.denormalize()
        outlist = []
        # Brute force for now; could use set instead...
        for data_a_elm in data_a_denorm:
            res = self._cmp_lists(data_a_elm, data_b_denorm)
            if is_or:
                res = not res

            if res:
                outlist.append(data_a_elm)

        return DictSam(outlist, True)

    def __getitem__(self, key):
        return self._data[key]

    def __sub__(self, other):
        return self._sub_and(other)

    def __and__(self, other):
        return self._sub_and(other, False)

    def __or__(self, other):
        return DictSam(self.denormalize() + other.denormalize(), True)

    def __eq__(self, other):
        return self.get_data() == other.get_data()

class JsonSam(DictSam):
    '''
    JSON split and merge (DICTSAM) main class.
    '''
    def __init__(self, fname, fname_aux=None, ignore_leaves=False, enforce_unique=False):
        (denormed_input, data) = self._load_data(fname)
        super().__init__(data, True, enforce_unique)
        self.donorm = denormed_input
        self.denorm_accum = data
        self.is_std = False
        if fname_aux and fname_aux[0]:
            self.json_sam_aux = JsonSam(fname_aux, enforce_unique=enforce_unique)

        self.set_ignore_leaves(ignore_leaves)

    def process(self, fname, fname_aux=None, outfile=None, set_op=None):
        '''
        Process a normalized or denormalized JSON file
        '''
        if fname[0] == STDIN and not outfile:
            self.is_std = True
            mkfn = lambda f, s: STDOUT
        elif outfile:
            mkfn = lambda f, s: outfile
            if outfile == STDOUT:
                self.is_std = True
        else:
            mkfn = lambda f, s: f.parent / Path(f.stem + s).with_suffix(f.suffix)

        if fname_aux[0]:
            ret = self._do_set_op(set_op)
            outpath = mkfn(fname[0], '-norm')
            self._write_normed(ret.get_data(), outpath)
        else:
            if set_op:
                raise NotImplementedError('Operand file (-F) required for "{}" operation'
                                          .format(set_op))
            if self.donorm:
                outpath = mkfn(fname[0], '-norm')
                self._write_normed(self.get_data(), outpath)
            else:
                outpath = mkfn(fname[0], '-denorm')
                self._write_denormed(outpath)

    def _do_set_op(self, set_op):
        if set_op == 'union':
            return self | self.json_sam_aux
        if set_op == 'except':
            return self - self.json_sam_aux
        if set_op == 'intersect':
            return self & self.json_sam_aux

        raise NotImplementedError('Invalid set operation "{}"'.format(set_op))

    def _load_data(self, files):
        '''
        Loads either a standard JSON file (normalized) or a mutable
        denormalized json file.  Denormalized files detected from header.

        path -- Input file name

        '''

        ret = []
        denormed_input = None
        for fname in files:
            with open(fname, 'r') as handle:
                raw_data = handle.read()
            try:
                data = json.loads(raw_data)
                if not isinstance(data, (dict, list)):
                    raise TypeError("{} must contain root of dictionary or list type"
                                    .format(fname))
                # Note this can alias a valid single-line path as a normalized input...
                ret += self.denormalize(data)
                if denormed_input is None:
                    denormed_input = False
                elif denormed_input is True:
                    raise TypeError("{} must be normalized consistent with other input files"
                                    .format(fname))
            except json.decoder.JSONDecodeError:
                for json_path in raw_data.split('\n'):
                    if json_path:
                        try:
                            ret.append(json.loads(json_path))
                        except json.decoder.JSONDecodeError:
                            # Allow denormed prefix for some edge cases (eg onepath)
                            ret.append(json.loads(json_path[1:]))

                if denormed_input is None:
                    denormed_input = True
                elif denormed_input is False:
                    raise TypeError("{} must be denormalized consistent with other input files"
                                    .format(fname))

        # Unfortunately results in lexicographic sort of integers because all
        # elms need to convert to strings to avoid "int < str" exceptions, and
        # python3 no longer allows passing a comparator function.
        return (denormed_input, sorted(ret, key=lambda a: [str(x) for x in a]))

    def _write_normed(self, data, outpath):
        ''' Write normalized json file to disk '''
        mixed_dict = data
        out_json = json.dumps(mixed_dict, indent=2)

        if not self.is_std and not outpath.suffix:
            outpath = outpath.with_suffix('.json')
        with open(outpath, 'w') as handle:
            handle.write(out_json)
        if not self.is_std:
            print("Updated JSON file written to {}".format(outpath))

    def _write_denormed(self, outpath):
        ''' Write denormalized json file to disk '''
        denormed = self.denorm_accum
        out_txt = ''
        for path in denormed:
            out_txt += json.dumps(path) + '\n'
        if len(out_txt) > 0:
            out_txt = out_txt[:-1]
        if len(denormed) == 1:
            # Disambiguate from valid single JSON input list
            out_txt = '_' + out_txt

        if not self.is_std and not outpath.suffix:
            outpath = outpath.with_suffix('.json')
        with open(outpath, 'w') as handle:
            handle.write(out_txt)
        if not self.is_std:
            print("Make edits to {} and then rerun with modified file to generate "
                  "updated JSON output"
                  .format(outpath))

def main():
    ''' CLI entry point '''
    arg_parser = argparse.ArgumentParser()
    group = arg_parser.add_mutually_exclusive_group()
    group.add_argument('-e', dest='set_op', required=False, default=None,
                       action='store_const', const='except',
                       help='Except (subtract/remove)')
    arg_parser.add_argument('-E', dest='enforce_unique', required=False,
                            action='store_true', default=False,
                            help='Enforce unique paths on normalization (no overwrite)')
    arg_parser.add_argument('-F', dest='infileaux', required=False,
                            default=None, type=Path,
                            help='Operand input JSON file')
    arg_parser.add_argument('-I', dest='ignore_leaves', required=False,
                            action='store_true', default=False,
                            help='Ignore leaf values for set operations')
    group.add_argument('-i', dest='set_op', required=False, default=None,
                       action='store_const', const='intersect',
                       help='Intersect')
    arg_parser.add_argument('-o', dest='outfile', required=False,
                            default=None, type=Path,
                            help='Output JSON file (autogenerated name if omitted)')
    group.add_argument('-u', dest='set_op', required=False, default=None,
                       action='store_const', const='union',
                       help='Union (add/merge)')
    arg_parser.add_argument('infiles', nargs='*', default=[STDIN],
                            help='Input files ("-" for last file outputs to stdout)')
    args = arg_parser.parse_args()

    infiles = [Path(x) for x in args.infiles]
    infileaux = [args.infileaux]

    if infiles[-1] == Path('-'):
        infiles.pop()
        outfile = STDOUT
    else:
        outfile = args.outfile

    json_sam = JsonSam(infiles, infileaux, args.ignore_leaves, args.enforce_unique)
    json_sam.process(infiles, infileaux, outfile, args.set_op)

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['jsonsam']

package_data = \
{'': ['*']}

install_requires = \
['Faker>=8.1,<9.0', 'deepdiff>=5.5,<6.0']

entry_points = \
{'console_scripts': ['jsonsam = jsonsam.jsonsam:main',
                     'randdict = jsonsam.randdict:main']}

setup_kwargs = {
    'name': 'jsonsam',
    'version': '0.1.1',
    'description': 'Tool to facilitate editing, splitting and merging of nested JSON and Python dictionaries',
    'long_description': '# JSONSAM Overview\n\n## Quickstart\n\nThe following is a minimal example of the JSONSAM CLI workflow:\n\n```console\n$ jsonsam my.json\nMake edits to my-denorm.json and then rerun with modified file to generate updated JSON output\n$ vim my-denorm.json \t\t# Add/delete/edit line-based paths\n$ jsonsam my-denorm.json \t# Convert edits back to standard JSON\nUpdated JSON file written to my-denorm-norm.json\n```\n\nThe finished product (the standard JSON file with edits applied) is\n`my-denorm-norm.json`. See the [Command Line Examples](#command-line-examples)\nsection for extensive additional examples.\n\n## Introduction\n\nJSONSAM captures some of the techniques I\'ve used over the years to slice and\ndice large chunks (> 1MB) of highly nested JSON.  vim is still my preferred\neditor and I like to use line-based shell primitives like sed and grep when\npossible.  However, since conventional JSON is not line-framed and\nJSON-streaming has multipath semantics, it is generally not amenable to direct\nmodification using conventional Unix utilities.\n\nJSONSAM addresses this by converting JSON into an intermediate denormalized\nformat (not to be confused with JSON-streaming or new-line delimited JSON).\nThe denormalized format captures each path in a JSON list of line-delimited\npaths.  These paths can then be edited with standard tools to add, delete,\nmodify, shift-up/down, reorder nested lists, etc.  Once desired edits are made,\nthe denormalized file is renormalized back into a standard JSON file.  JSONSAM\ncan also work on entire directories of files at once, allowing them to be\neasily merged, split and modified.  Denormalized form might also be useful for\nstreaming if one is willing to trade data expansion for framing.\n\nJSONSAM can also perform set arithmetic (union, intersection and difference)\nagainst a pair of JSON input files.  This is useful when a template or\nreference file exists and it\'s desired to extract, remove or augment another\nJSON file based on this template.\n\nLastly, a random realistic dictionary generator is included as part of the test\ninfrastructure. It creates nested dictionaries and lists of random breadth and\nwidth.\n\nJSONSAM comprises the following components:\n\n- jsonsam: Command line utility for processing JSON\n- randdict: Command line utility for generating random dictionaries\n- JsonSam: Denormalize, edit, normalize and perform set arithmetic on JSON files\n- DictSam: Denormalize, edit, normalize and perform set arithmetic on nested dictionaries/lists\n- DictGen: Generate random deeply nested dictionaries and lists with realistic data\n\n## Installation\n\n### Install from PyPi:\n\n`pip install jsonsam`\n\n### Importing\n\n```python\n>>> from jsonsam import DictSam # Dictionary denorm, norm, and set arithmetic\n>>> from jsonsam import DictGen # Random realistic nested dictionary/list generator\n```\n\n# Command Line Examples\n\n```console\n# online help\n$ jsonsam -h\n```\n\nConsider the following file `simple.json`:\n```json\n$ cat simple.json\n{\n  "Bad": true,\n  "beer": 3.14,\n  "rots": "Our",\n  "young": {\n    "guts": "but",\n    "Vodka": [\n      27,\n      [\n        "goes",\n        2.718,\n        {\n          "Well": "water",\n          "max": "plank"\n        },\n        "old"\n      ]\n    ],\n    "bark": null,\n    "last": 66\n  }\n}\n```\nThis file is in normalized form.  First, we denormalize the file:\n\n```console\n$ jsonsam simple.json\nMake edits to simple-denorm.json and then rerun with modified file to generate updated JSON output\n```\n\nThe denormalized file, `simple-denorm.json`, now contains a list of paths.\nEach line represents a path from root to leaf through the original JSON file\nwith non-terminal integers representing list indices:\n\n```json\n$ cat simple-denorm.json\n["Bad", true]\n["beer", 3.14]\n["rots", "Our"]\n["young", "Vodka", 0, 27]\n["young", "Vodka", 1, 0, "goes"]\n["young", "Vodka", 1, 1, 2.718]\n["young", "Vodka", 1, 2, "Well", "water"]\n["young", "Vodka", 1, 2, "max", "plank"]\n["young", "Vodka", 1, 3, "old"]\n["young", "bark", null]\n["young", "guts", "but"]\n["young", "last", 66]\n```\n\nAll other rows can be arbitrarily mutated, added, removed and reordered\n(non-terminal integers denote nested list indices).  When the desired result is\nachieved, rerun `jsonsam` with the edited `simple-denorm.json` as input to\ngenerate a renormalized `simple-denorm-norm.json` output file.  This represents\nthe original JSON file with the desired additions, deletions and mutations\napplied.  Note that JSONSAM appends `-norm` and `-denorm` strings to the file\nstem for convenience to avoid overwriting original files and track lineage.\nThis behavior can be overridden with the `-o <outputfile>` option.\n\n**Unless otherwise noted all examples that follow will start from `simple.json`\nand `simple-denorm.json`.**\n\n## Relocate dictionary slice from nested lists\nSuppose we want to shift-up the pair `"Well": "water"` out of the dictionary\nnested in the list of lists.  This can be achieved by applying the following\nsingle-line diff to `simple-denorm.json`:\n\n```diff\n ["young", "Vodka", 0, 27]\n ["young", "Vodka", 1, 0, "goes"]\n ["young", "Vodka", 1, 1, 2.718]\n-["young", "Vodka", 1, 2, "Well", "water"]\n+["young", "Vodka", 2, "Well", "water"]\n ["young", "Vodka", 1, 2, "max", "plank"]\n ["young", "Vodka", 1, 3, "old"]\n ["young", "bark", null]\n```\n\nTo apply the changes, we normalize the path list by rerunning JSONSAM:\n\n```diff\n$ jsonsam simple-denorm.json \nUpdated JSON file written to simple-denorm-norm.json\n$ diff --unified=20 <(cat simple.json |jq --sort-keys \'.\') <(cat simple-denorm-norm.json |jq --sort-keys \'.\')\n {\n   "Bad": true,\n   "beer": 3.14,\n   "rots": "Our",\n   "young": {\n     "Vodka": [\n       27,\n       [\n         "goes",\n         2.718,\n         {\n-          "Well": "water",\n           "max": "plank"\n         },\n         "old"\n-      ]\n+      ],\n+      {\n+        "Well": "water"\n+      }\n     ],\n     "bark": null,\n     "guts": "but",\n     "last": 66\n   }\n }\n```\n\nWe see in the above JSON that indeed the `"Well": "water"` pair has been\nplucked from the nested dictionary, and a new dictionary containing the pair\nhas been appended to the outer list.  `simple-denorm-norm.json` now represents\nthe finished product.\n\n## Click to expand additional examples:\n\n<details><summary>Remove an item, and convert a leaf string to a dictionary</summary>\n\n## Remove an item, and convert a leaf string to a dictionary\n\nIn this case we do two things:\n\n1. Remove the `young.bark.null` path\n1. Convert `young.guts.but` path from a terminal string to a dictionary of\n`young.guts.{but:dog, route:cat}`\n\nThis can be achieved by applying the following diff to `simple-denorm.json`:\n\n```json\n ["young", "Vodka", 1, 2, "Well", "water"]\n ["young", "Vodka", 1, 2, "max", "plank"]\n ["young", "Vodka", 1, 3, "old"]\n-["young", "bark", null]\n-["young", "guts", "but"]\n+["young", "guts", "but", "dog"]\n+["young", "guts", "route", "cat"]\n ["young", "last", 66]\n```\n\nLooking at `simple-denorm-norm.json` we see `young.bark.null` has been removed\nand we\'ve converted `young.guts.but` from a terminal string into a dictionary\nof `{but:dog, route:cat}`:\n\n```json\n$ jsonsam simple-denorm.json\nUpdated JSON file written to simple-denorm-norm.json\n$ cat simple-denorm-norm.json\n{\n  "Bad": true,\n  "beer": 3.14,\n  "rots": "Our",\n  "young": {\n    "Vodka": [\n      27,\n      [\n        "goes",\n        2.718,\n        {\n          "Well": "water",\n          "max": "plank"\n        },\n        "old"\n      ]\n    ],\n    "guts": {\n      "but": "dog",\n      "route": "cat"\n    },\n    "last": 66\n  }\n}\n```\n</details>\n\n<details><summary>Convert a terminal string to a list of dictionaries</summary>\n\n## Convert a terminal string to a list of dictionaries\n\nHere we convert the `young.guts` path to a list of dictionaries.\nThis can be achieved by applying the following diff to `simple-denorm.json`:\n\n```json\n ["young", "Vodka", 1, 2, "max", "plank"]\n ["young", "Vodka", 1, 3, "old"]\n ["young", "bark", null]\n-["young", "guts", "but"]\n+["young", "guts", 0, "but", "dog"]\n+["young", "guts", 1, "route", "dog"]\n ["young", "last", 66]\n```\n\nAnd the resulting normalized JSON:\n\n```json\n$ jsonsam simple-denorm.json\nUpdated JSON file written to simple-denorm-norm.json\n$ cat simple-denorm-norm.json\n{\n  "Bad": true,\n  "beer": 3.14,\n  "rots": "Our",\n  "young": {\n    "Vodka": [\n      27,\n      [\n        "goes",\n        2.718,\n        {\n          "Well": "water",\n          "max": "plank"\n        },\n        "old"\n      ]\n    ],\n    "bark": null,\n    "guts": [\n      {\n        "but": "dog"\n      },\n      {\n        "route": "dog"\n      }\n    ],\n    "last": 66\n  }\n}\n```\n\n</details>\n\n\n<details><summary>Add path and reorder a list</summary>\n\n### Add path and reorder a list\n\nThree tasks here:\n1. Add a new `jack.beanstock.happy.941` path\n1. Move list item `young.Vodka.1.2` to `young.Vodka.2`\n1. Add the string "append" to `young.Vodka` list\n\n\nThis can be achieved by applying the following diff to `simple-denorm.json`:\n\n```json\n ["Bad", true]\n ["beer", 3.14]\n ["rots", "Our"]\n+["jack", "beanstock", "happy", 941]\n ["young", "Vodka", 0, 27]\n-["young", "Vodka", 1, 0, "goes"]\n-["young", "Vodka", 1, 1, 2.718]\n-["young", "Vodka", 1, 2, "Well", "water"]\n-["young", "Vodka", 1, 2, "max", "plank"]\n+["young", "Vodka", 1, 1, "goes"]\n+["young", "Vodka", 1, 0, 2.718]\n+["young", "Vodka", 2, "Well", "water"]\n+["young", "Vodka", 2, "max", "plank"]\n+["young", "Vodka", 3, "appended"]\n ["young", "Vodka", 1, 3, "old"]\n ["young", "bark", null]\n ["young", "guts", "but"]\n```\n\nWhich results in:\n\n```json\n$ jsonsam simple-denorm.json\nUpdated JSON file written to simple-denorm-norm.json\n$ cat simple-denorm-norm.json\n{\n  "Bad": true,\n  "beer": 3.14,\n  "jack": {\n    "beanstock": {\n      "happy": 941\n    }\n  },\n  "rots": "Our",\n  "young": {\n    "Vodka": [\n      27,\n      [\n        2.718,\n        "goes",\n        null,\n        "old"\n      ],\n      {\n        "Well": "water",\n        "max": "plank"\n      },\n      "appended"\n    ],\n    "bark": null,\n    "guts": "but",\n    "last": 66\n  }\n}\n```\n\nRemoved list elements become `null` to preserve the ordinal of retained\nelements.\n</details>\n\n<details><summary>Unix pipeline processing</summary>\n\n## Unix pipeline processing\n\nIf no input files are provided, jsonsam defaults to reading from stdin and\nwriting to stdout.  This allows use of standard Unix pipelines for editing JSON\nfiles with line-based shell utilities like sed, grep, awk, etc.\n\nIn the below example, we use grep to remove all `Vodka` keys and sed to convert `beer` to `Whiskey`:\n\n```json\n$ cat simple.json |jsonsam |grep -v Vodka |sed \'s/beer/Whiskey/g\' |jsonsam\n{\n  "Bad": true,\n  "Whiskey": 3.14,\n  "rots": "Our",\n  "young": {\n    "bark": null,\n    "guts": "but",\n    "last": 66\n  }\n}\n```\n\n</details>\n\n<details><summary>Transmogrify 7 JSON files into 11 JSON files, round-robin paths</summary>\n\n### Transmogrify 7 JSON files into 11 JSON files, round-robin paths\n\nIn this example, we will generate 7 random JSON files.  Next we will merge them\nand then split the merge into 11 files, round-robinning the paths across the 11\nnew JSON files:\n\n```console\n$ randdict -B5 -D5 -c 7 -o data.json\n$ ls\ndata000.json  data001.json  data002.json  data003.json  data004.json  data005.json  data006.json\n$ jsonsam *.json - |split -xn r/11\n$ ls\ndata000.json  data001.json  data002.json  data003.json  data004.json  data005.json  data006.json  x00  x01  x02  x03  x04  x05  x06  x07  x08  x09  x0a\n$ ls x* |xargs -IF jsonsam F\nUpdated JSON file written to x00-norm.json\nUpdated JSON file written to x01-norm.json\nUpdated JSON file written to x02-norm.json\nUpdated JSON file written to x03-norm.json\nUpdated JSON file written to x04-norm.json\nUpdated JSON file written to x05-norm.json\nUpdated JSON file written to x06-norm.json\nUpdated JSON file written to x07-norm.json\nUpdated JSON file written to x08-norm.json\nUpdated JSON file written to x09-norm.json\nUpdated JSON file written to x0a-norm.json\n```\n\nThe 11 `x<num>-norm.json` files now contain all the data in the original 7\nfiles, but split round-robin across 11 files.  Substituting `split -xn l/11`\nfor the split command would instead do a conventional partition rather than\nround-robin.  Of course any intermediate mutations, additions and deletions\ncould also be made in the same pipeline.\n\n</details>\n\n## Comand Line Set Operations Examples\n\nSet operations union `|`, intersection `&` and difference `-` can be performed\non JSON files specified with positional arguments (lhs) and `-F` (rhs) options.\nThe following examples will use the original `simple.json` file as the\nleft-hand operand and `simple_aux.json` as the right-hand operand:\n\n```json\n$ cat simple_aux.json\n{\n  "Bad": true,\n  "beer": 3.1415,\n  "NEW": "PATH",\n  "young": {\n    "Vodka": [\n      27\n    ],\n    "bark": null,\n    "guts": "but",\n    "last": 66\n  }\n}\n```\n\n## Difference/except `simple.json - simple_aux.json`\n\nIn the following example we perform the set difference of `simple.json -\nsimple_aux.json`.  Paths common to both `simple.json` and `simple_aux.json` are\nremoved from `simple.json`.\n\nNote the `-I` option can also be added for all operations to ignore leaf values\nin comparison.  This is useful when one wishes to use an existing JSON file as\na template for paths to remove regardless of value.\n\nAlso note removed list elements become `null` to preserve the ordinal of\nretained elements.\n\n```diff\n$ jsonsam simple.json -F simple_aux.json -e\nUpdated JSON file written to simple-norm.json\n$ diff --unified=20 <(cat simple.json |jq --sort-keys \'.\') <(cat simple-norm.json |jq --sort-keys \'.\')\n {\n-  "Bad": true,\n   "beer": 3.14,\n   "rots": "Our",\n   "young": {\n     "Vodka": [\n-      27,\n+      null,\n       [\n         "goes",\n         2.718,\n         {\n           "Well": "water",\n           "max": "plank"\n         },\n         "old"\n       ]\n-    ],\n-    "bark": null,\n-    "guts": "but",\n-    "last": 66\n+    ]\n   }\n }\n```\n\n## Click to expand additional set examples:\n\n<details><summary>Intersection (simple.json & simple_aux.json)</summary>\n\n## Intersection `(simple.json & simple_aux.json)`\n\nIn this example we perform the intersection of the two JSON files.  Note that\n`-I` is passed, so the leaf values are ignored allowing us to also intersect\nwith a slightly less precise value of pi.\n\n```json\n$ jsonsam simple.json -F simple_aux.json -iI\nUpdated JSON file written to simple-norm.json\n$ cat simple-norm.json\n{\n  "Bad": true,\n  "beer": 3.14,\n  "young": {\n    "Vodka": [\n      27\n    ],\n    "bark": null,\n    "guts": "but",\n    "last": 66\n  }\n}\n```\n\n</details>\n\n<details><summary> Union/merge (simple.json | simple_aux.json)</summary>\n\n## Union/merge `(simple.json | simple_aux.json)`\n\nIn this example we perform the union of the two JSON files.  Note the\nconflicting `beer` key is overwritten by the right-hand side\'s value and the\n`NEW` item has been added.\n\n```json\n$ jsonsam simple.json -F simple_aux.json -u\nUpdated JSON file written to simple-norm.json\n$ cat simple-norm.json\n{\n  "Bad": true,\n  "beer": 3.1415,\n  "rots": "Our",\n  "young": {\n    "Vodka": [\n      27,\n      [\n        "goes",\n        2.718,\n        {\n          "Well": "water",\n          "max": "plank"\n        },\n        "old"\n      ]\n    ],\n    "bark": null,\n    "guts": "but",\n    "last": 66\n  },\n  "NEW": "PATH"\n}\n```\n\n</details>\n\n## Random nested dictionary generator examples\n\nA random dictionary generator is included in the package.  It generates deeply\nnested dictionaries of realistic random data.  Random nested lists are also\nsupported.  The full range of JSON types are supported.  Each step in breadth\nand depth is uniquely randomized.\n\nTo generate a random dictionary of max breadth 3, max depth 4 consisting of 40%\nnested lists (compact form for brevity):\n\n```json\n$ randdict -B 3 -D 4 -l 40 |jq -c \'.\'\n{"game":{"yourself":{"difficult":"0181 Aguilar Parkways","always":{"before":530982.9961689024,"leader":{"kid":0.41913904357146525,"available":523048.62318369857},"effect":-609888.9269447384}},"believe":[[{"ok":0.24842658485754932,"kind":"91390 Williams Forges Apt. 824","charge":0.7294452894392176},"482 Bonnie Route",{"cup":null,"hotel":-292649,"bill":0.6952953662736593}],[0.47409833741964447,"093 Becker Meadow"]]},"standard":[{"former":null,"make":270989.99712404597,"pay":245327},{"couple":"Ward-Wright","say":{"state":["Miller LLC","Simpson, Cooper and Cole","Baird, Wilson and Barnes"],"though":"76018 Peterson Keys Suite 971","better":0.28193072232673766},"you":{"expect":null,"matter":{"hear":"49746 Johnson Mountain","artist":"Carr Group","interest":"2366 Miller Mission"}}},{"often":{"reason":"68465 Rosario Drive Apt. 223","early":{"actually":null,"memory":null,"theory":null},"nor":"White, Campbell and Thomas"},"hope":{"poor":2548096.1722068526,"others":372384,"from":["Vaughn PLC","Pena, Singh and Bryant",322924]}}],"truth":[{"state":{"two":{"response":0.08053812548862638,"success":null,"protect":"Long, Ross and Garcia"},"institution":null,"assume":1888192},"cut":[null,{"audience":null,"grow":"Alexander Inc","blood":"Williams, Smith and Hernandez"},1358976.5609615073],"himself":["39624 Guzman Mountains",-2605707.399841271]},["Wilson Group",0.2127797923458118,{"man":{"media":null,"member":725008.2758262581},"forget":1022493.1525128945,"guess":"Thomas Ltd"}]]}\n```\n\n# API Example\n\nThe following example shows how to use the API to generate a random test\ndictionary, pick two overlapping subsets of the dictionary, then take the\ndifference, union and intersection of these dictionaries:\n\n```python\n>>> from jsonsam import DictSam\n>>> from jsonsam import DictGen\n>>> test_dict = DictGen().gen_fake_dict(breadth_rng=(2, 4), depth_rng=(3, 5), list_dist=(3, 1))\n>>> dict_sam_orig = DictSam(test_dict)\n>>> dict_sam_a = dict_sam_orig.random_dict_pick(60)\n>>> dict_sam_b = dict_sam_orig.random_dict_pick(80)\n>>> diff = dict_sam_b - dict_sam_a\n>>> union = dict_sam_b | dict_sam_a\n>>> intersect = dict_sam_b & dict_sam_a\n```\n\n**NOTE: At the moment only JSON-serializable structures are supported by DictSam**\n\nFull API readthedocs coming soon...\n\n# Authors\n\nSee [AUTHORS](AUTHORS.md)\n',
    'author': 'Eric D. Cohen',
    'author_email': 'os@epieye.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/project705/jsonsam',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
